#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""This module defines writer and reader classes for a fast, immutable
on-disk key-value database format. The current format is identical
to D. J. Bernstein's CDB format (http://cr.yp.to/cdb.html).
"""

from array import array
from bisect import bisect_right
from collections import defaultdict
from cPickle import dumps, loads
from struct import pack, unpack, calcsize


def cdb_hash(key):
    h = 5381L
    for c in key:
        h = (h + (h << 5)) & 0xffffffffL ^ ord(c)
    return h

# Read/write convenience functions

def readint(map, offset):
    return unpack("<L", map[offset:offset+4])[0]

def readints(map, offset, length):
    return unpack("<" + "L" * length, map[offset: offset + 4 * length])

def writeints(f, value1, value2):
    f.write(pack("<LL", value1, value2))

# Key/value encoding/decoding functions

def encode_key(term):
    fieldnum, text = term
    return ("%03X" % fieldnum) + text.encode("utf8")

def decode_key(key):
    return (int(key[:3], 16), key[3:].decode("utf8"))

def encode_docnum(docnum):
    return "%08X" % docnum

def decode_docnum(key):
    return int(key, 16)

def encode_vectorkey(docnum_and_fieldnum):
    return "%08X%03X" % docnum_and_fieldnum

def decode_vectorkey(key):
    return (int(key[:8], 16), int(key[8:], 16))

def enpickle(data):
    return dumps(data, -1)

depickle = loads

# Convenience function to configure TableWriter/Reader objects with the
# proper key encoding/decoding functions for various specific jobs.

def create_term_table(storage, segment):
    return storage.create_posting_table(segment.term_filename,
                                        segment.posts_filename,
                                        keycoder=encode_key,
                                        valuecoder=enpickle)
    
def create_docs_table(storage, segment):
    return storage.create_list(segment.docs_filename,
                               valuecoder=enpickle)

def create_vector_table(storage, segment):
    return storage.create_posting_table(segment.vector_filename,
                                        segment.vectorposts_filename,
                                        keycoder=encode_vectorkey,
                                        valuecoder=enpickle,
                                        stringids = True)

def open_term_table(storage, segment):
    return storage.open_posting_table(segment.term_filename,
                                      segment.posts_filename,
                                      keycoder=encode_key,
                                      keydecoder=decode_key,
                                      valuedecoder=depickle)

def open_docs_table(storage, segment, schema):
    storedfieldnames = schema.stored_field_names()
    def dictifier(value):
        value = loads(value)
        return dict(zip(storedfieldnames, value))
    return storage.open_list(segment.docs_filename,
                             segment.doc_count_all(),
                             valuedecoder=dictifier)
    
def open_vector_table(storage, segment):
    return storage.open_posting_table(segment.vector_filename,
                                      segment.vectorposts_filename,
                                      keycoder=encode_vectorkey,
                                      keydecoder=decode_vectorkey,
                                      valuedecoder=depickle,
                                      stringids=True)
    

# Used top copy vector postings directly from one table to another during
# segment merging, since the information in the postings doesn't need to be
# updated.

def copy_postings(treader, inkey, twriter, outkey, buffersize=32*1024):
    offset, length, postcount, data = treader._get(inkey)
    twriter.add(outkey, data,
                postinginfo=(twriter.offset, length, postcount))
    
    # Copy the raw posting data
    infile = treader.postingfile
    outfile = twriter.postingfile
    if length <= buffersize:
        outfile.write(infile.map[offset: length])
    else:
        infile.seek(offset)
        sofar = 0
        while sofar < length:
            readsize = min(buffersize, length - sofar)
            outfile.write(infile.read(readsize))
            sofar += readsize
    
    twriter.offset = outfile.tell()


# Table classes

class FileHashWriter(object):
    def __init__(self, dbfile):
        self.dbfile = dbfile
        dbfile.seek(2048)
        self.hashes = defaultdict(list)
    
    def add_all(self, items):
        dbfile = self.dbfile
        hashes = self.hashes
        pos = dbfile.tell()
        write = dbfile.write
        
        for key, value in items:
            writeints(dbfile, len(key), len(value))
            write(key + value)
            
            h = cdb_hash(key)
            hashes[h & 255].append((h, pos))
            pos += len(key) + len(value) + 8
    
    def add(self, key, value):
        self.add_all(((key, value), ))
    
    def _write_hashes(self):
        dbfile = self.dbfile
        hashes = self.hashes
        directory = self.directory = []
        
        pos = dbfile.tell()
        for i in xrange(0, 256):
            entries = hashes[i]
            numslots = 2 * len(entries)
            directory.append((pos, numslots))
            
            null = (0, 0)
            hashtable = [null] * numslots
            for hashval, position in entries:
                n = (hashval >> 8) % numslots
                while hashtable[n] is not null:
                    n = (n + 1) % numslots
                hashtable[n] = (hashval, position)
            
            for hashval, position in hashtable:
                writeints(dbfile, hashval, position)
                pos += 8
                
        dbfile.flush()
    
    def _write_directory(self):
        dbfile = self.dbfile
        directory = self.directory
        
        dbfile.seek(0)
        for position, numslots in directory:
            writeints(dbfile, position, numslots)
        assert dbfile.tell() == 2048
        dbfile.flush()
    
    def close(self):
        self._write_hashes()
        self._write_directory()
        self.dbfile.close()


class FileHashReader(object):
    def __init__(self, dbfile):
        self.dbfile = dbfile
        self.map = dbfile.map
        self.end_of_data = unpack("<L", self.map[0:4])[0]
        self.is_closed = False
        
    def close(self):
        if self.is_closed:
            raise Exception("Tried to close %r twice" % self)
        del self.map
        self.dbfile.close()
        self.is_closed = True
        
    def read(self, position, length):
        return self.map[position:position+length]

    def read2ints(self, position):
        return unpack("<LL", self.map[position:position+8])

    def _ranges(self, pos=2048):
        read2ints = self.read2ints
        eod = self.end_of_data
        
        while pos < eod:
            keylen, datalen = read2ints(pos)
            keypos = pos + 8
            datapos = pos + 8 + keylen
            pos = datapos + datalen
            yield (keypos, keylen, datapos, datalen)

    def __iter__(self):
        return self.items()

    def items(self):
        read = self.read
        for keypos, keylen, datapos, datalen in self._ranges():
            yield (read(keypos, keylen), read(datapos, datalen))

    def keys(self):
        read = self.read
        for keypos, keylen, _, _ in self._ranges():
            yield read(keypos, keylen)

    def values(self):
        read = self.read
        for _, _, datapos, datalen in self._ranges():
            yield read(datapos, datalen)

    def __getitem__(self, key):
        for data in self.all(key):
            return data
        raise KeyError(key)

    def get(self, key, default=None):
        for data in self.all(key):
            return data
        return default

    def _hashtable_info(self, keyhash):
        return self.read2ints(keyhash << 3 & 2047)

    def _key_position(self, key):
        keyhash = cdb_hash(key)
        hpos, hslots = self._hashtable_info(keyhash)
        if not hslots:
            raise KeyError(key)
        slotpos = hpos + (((keyhash >> 8) % hslots) << 3)
        u, pos = self.read2ints(slotpos)
        return pos

    def _get_ranges(self, key):
        read = self.read
        read2ints = self.read2ints
        keyhash = cdb_hash(key)
        hpos, hslots = self._hashtable_info(keyhash)
        if not hslots:
            return
        
        slotpos = hpos + (((keyhash >> 8) % hslots) << 3)
        for _ in xrange(0, hslots):
            u, pos = read2ints(slotpos)
            if not pos:
                return
            
            slotpos += 8
            # If we reach the end of the hashtable, wrap around
            if slotpos == hpos + (hslots << 3):
                slotpos = hpos
            
            if u == keyhash:
                keylen, datalen = read2ints(pos)
                if keylen == len(key):
                    if key == read(pos + 8, keylen):
                        yield (pos + 8 + keylen, datalen)
        
    def all(self, key):
        read = self.read
        for datapos, datalen in self._get_ranges(key):
            yield read(datapos, datalen)
            
    def __contains__(self, key):
        for _ in self._get_ranges(key):
            return True
        return False


class OrderedHashWriter(FileHashWriter):
    def __init__(self, dbfile, blocksize = 100):
        super(OrderedHashWriter, self).__init__(dbfile)
        self.blocksize = blocksize
        self.index = []
        self.indexcount = None
        self.lastkey = None
        
    def add_all(self, items):
        dbfile = self.dbfile
        hashes = self.hashes
        pos = dbfile.tell()
        write = dbfile.write
        
        ix = self.index
        ic = self.indexcount
        bs = self.blocksize
        lk = self.lastkey
        
        for key, value in items:
            if key <= lk:
                raise ValueError("Keys must increase: %r .. %r" % (lk, key))
            lk = key
            
            if ic is None:
                ix.append(key)
                ic = 0
            else:
                ic += 1
                if ic == bs:
                    ix.append(key)
                    ic = 0
            
            writeints(dbfile, len(key), len(value))
            write(key + value)
            
            h = cdb_hash(key)
            hashes[h & 255].append((h, pos))
            pos += len(key) + len(value) + 8
        
        self.indexcount = ic
        self.lastkey = lk
        
    def close(self):
        self._write_hashes()
        self.dbfile.write_pickle(self.index)
        self._write_directory()
        self.dbfile.close()

      
class OrderedHashReader(FileHashReader):
    def __init__(self, dbfile):
        super(OrderedHashReader, self).__init__(dbfile)
        lastpos, lastnum = self.read2ints(255*8)
        dbfile.seek(lastpos + lastnum * 8)
        self.index = dbfile.read_pickle()

    def _closest_key(self, key):
        index = self.index
        i = max(0, bisect_right(index, key) - 1)
        return index[i]
    
    def _ranges_from(self, key):
        read = self.read
        ckey = self._closest_key(key)
        pos = self._key_position(ckey)
        
        if ckey != key:
            for keypos, keylen, _, _ in self._ranges(pos=pos):
                k = read(keypos, keylen)
                if k >= key:
                    pos = keypos - 8
                    break
        
        return self._ranges(pos=pos)

    def items_from(self, key):
        read = self.read
        for keypos, keylen, datapos, datalen in self._ranges_from(key):
            yield (read(keypos, keylen), read(datapos, datalen))

    def keys_from(self, key):
        read = self.read
        for keypos, keylen, _, _ in self._ranges_from(key):
            yield read(keypos, keylen)

    def values(self, key):
        read = self.read
        for _, _, datapos, datalen in self._ranges_from(key):
            yield read(datapos, datalen)


class FileTableWriter(OrderedHashWriter):
    def __init__(self, dbfile, keycoder=None, valuecoder=None):
        sup = super(FileTableWriter, self)
        sup.__init__(dbfile)
        self.keycoder = keycoder or str
        self.valuecoder = valuecoder or enpickle
        
        self._add = sup.add
        
    def add(self, key, data):
        key = self.keycoder(key)
        data = self.valuecoder(data)
        self._add(key, data)
        

class FileTableReader(OrderedHashReader):
    def __init__(self, dbfile, keycoder=None, keydecoder=None, valuedecoder=None):
        sup = super(FileTableReader, self)
        sup.__init__(dbfile)
        self.keycoder = keycoder or str
        self.keydecoder = keydecoder or int
        self.valuedecoder = valuedecoder or depickle
        
        self._items = sup.items
        self._items_from = sup.items_from
        self._keys = sup.keys
        self._keys_from = sup.keys_from
        
    def get(self, key):
        k = self.keycoder(key)
        return self.valuedecoder(self[k])
    
    def __contains__(self, key):
        k = self.keycoder(key)
        return super(FileTableReader, self).__contains__(k)
    
    def items(self):
        kd = self.keydecoder
        vd = self.valuedecoder
        for key, value in self._items():
            yield (kd(key), vd(value))
            
    def items_from(self, key):
        fromkey = self.keycoder(key)
        kd = self.keydecoder
        vd = self.valuedecoder
        for key, value in self._items_from(fromkey):
            yield (kd(key), vd(value))
            
    def keys(self):
        kd = self.keydecoder
        for k in self._keys():
            yield kd(k)
            
    def keys_from(self, key):
        kd = self.keydecoder
        for k in self._keys_from(self.keycoder(key)):
            yield kd(k)
            

class FilePostingTableWriter(FileTableWriter):
    def __init__(self, dbfile, postingfile, keycoder=None, valuecoder=None,
                 stringids=False):
        super(FilePostingTableWriter, self).__init__(dbfile, keycoder=keycoder, valuecoder=valuecoder)
        self.postingfile = postingfile
        self.stringids = stringids
        self.lastpostid = None
        self.postcount = 0
        self.offset = 0
        
    def close(self):
        super(FilePostingTableWriter, self).close()
        self.postingfile.close()
        
    def write_posting(self, id, data, writefn):
        if id <= self.lastpostid:
            raise IndexError("IDs must increase: %r..%r" % (self.lastpostid, id))
        
        pf = self.postingfile
        if self.stringids:
            pf.write_string(id.encode("utf8"))
        else:
            lastpostid = self.lastpostid or 0
            pf.write_varint(id - lastpostid)
        
        self.lastpostid = id
        self.postcount += 1
        
        return writefn(pf, data)
    
    def add(self, key, data, postinginfo=None):
        if postinginfo:
            offset, length, postcount = postinginfo
            endoffset = offset + length
        else:
            offset = self.offset
            endoffset = self.postingfile.tell()
            length = endoffset - offset
            postcount = self.postcount
        
        super(FilePostingTableWriter, self).add(key, (offset, length, postcount, data))
        
        # Reset the posting variables
        self.offset = endoffset
        self.postcount = 0
        self.lastpostid = None
    

class FilePostingTableReader(FileTableReader):
    def __init__(self, dbfile, postingfile, keycoder=None, keydecoder=None, valuedecoder=None,
                 stringids=False):
        sup = super(FilePostingTableReader, self)
        sup.__init__(dbfile, keycoder=keycoder, keydecoder=keydecoder,
                     valuedecoder=valuedecoder)
        self.postingfile = postingfile
        self.stringids = stringids
        
        self._get = sup.get
        if self.stringids:
            self._read_id = self._read_id_string
        else:
            self._read_id = self._read_id_varint
    
    def close(self):
        super(FilePostingTableReader, self).close()
        self.postingfile.close()
    
    def _read_id_varint(self, lastid):
        return lastid + self.postingfile.read_varint()
    
    def _read_id_string(self, lastid):
        return self.postingfile.read_string().decode("utf8")
    
    def get(self, key):
        return self._get(key)[3]
    
    def items(self):
        kd = self.keydecoder
        vd = self.valuedecoder
        for key, value in self._items():
            yield (kd(key), vd(value)[3])
            
    def items_from(self, key):
        fromkey = self.keycoder(key)
        kd = self.keydecoder
        vd = self.valuedecoder
        for key, value in self._items_from(fromkey):
            yield (kd(key), vd(value)[3])
    
    def posting_count(self, key):
        return super(FilePostingTableReader, self).get(key)[2]
    
    def _seek_postings(self, key):
        offset, _, count = self._get(key)[:3]
        self.postingfile.seek(offset)
        return count
    
    def postings(self, key, readfn):
        postingfile = self.postingfile
        _read_id = self._read_id
        id = 0
        for _ in xrange(0, self._seek_postings(key)):
            id = _read_id(id)
            yield (id, readfn(postingfile))


class FileRecordWriter(object):
    def __init__(self, dbfile, format):
        self.dbfile = dbfile
        self.format = format
        
    def close(self):
        self.dbfile.close()
        
    def append(self, args):
        self.dbfile.write(pack(self.format, *args))


class FileRecordReader(object):
    def __init__(self, dbfile, format):
        self.dbfile = dbfile
        self.map = dbfile.map
        self.format = format
        self.itemsize = calcsize(format)
    
    def close(self):
        del self.map
        self.dbfile.close()
    
    def get_record(self, recordnum):
        itemsize = self.itemsize
        return unpack(self.format, self.map[recordnum * itemsize:recordnum * itemsize + itemsize])
    
    def get(self, recordnum, itemnum):
        return self.get_record(recordnum)[itemnum]


class FileListWriter(object):
    def __init__(self, dbfile, valuecoder=str):
        self.dbfile = dbfile
        self.positions = array("L")
        self.lengths = array("L")
        dbfile.write_ulong(0)
        self.valuecoder = valuecoder
    
    def close(self):
        f = self.dbfile
        directory_pos = f.tell()
        f.write_array(self.positions)
        f.write_array(self.lengths)
        f.flush()
        f.seek(0)
        f.write_ulong(directory_pos)
        f.close()
    
    def append(self, value):
        f = self.dbfile
        self.positions.append(f.tell())
        v = self.valuecoder(value)
        self.lengths.append(len(v))
        f.write(v)
        

class FileListReader(object):
    def __init__(self, dbfile, length, valuedecoder=str):
        self.dbfile = dbfile
        self.length = length
        self.valuedecoder = valuedecoder
        
        offset = dbfile.get_ulong(0)
        dbfile.seek(offset)
        self.positions = dbfile.read_array("L", length)
        self.lengths = dbfile.read_array("L", length)
    
    def close(self):
        self.dbfile.close()
    
    def get(self, num):
        position = self.positions[num]
        length = self.lengths[num]
        v = self.dbfile.map[position:position+length]
        return self.valuedecoder(v)


# Utility functions

def dump_hash(hashreader):
    dbfile = hashreader.dbfile
    read = hashreader.read
    read2ints = hashreader.read2ints
    eod = hashreader.end_of_data
    
    # Dump hashtables
    for bucketnum in xrange(0, 255):
        pos, numslots = read2ints(bucketnum * 8)
        if numslots:
            print "Bucket %d: %d slots" % (bucketnum, numslots)
            
            dbfile.seek(pos)
            for j in xrange(0, numslots):
                print "  %X : %d" % read2ints(pos)
                pos += 8
    
    # Dump keys and values
    print "-----"
    dbfile.seek(2048)
    pos = 2048
    while pos < eod:
        keylen, datalen = read2ints(pos)
        keypos = pos + 8
        datapos = pos + 8 + keylen
        key = read(keypos, keylen)
        data = read(datapos, datalen)
        print "%d +%d,%d:%r->%r" % (pos, keylen, datalen, key, data)
        pos = datapos + datalen
        



    


import unittest

from whoosh import analysis, fields
from whoosh.filedb.filestore import RamStorage
from whoosh.filedb.filewriting import NO_MERGE

class TestReading(unittest.TestCase):
    def _create_index(self):
        s = fields.Schema(f1 = fields.KEYWORD(stored = True),
                          f2 = fields.KEYWORD,
                          f3 = fields.KEYWORD)
        st = RamStorage()
        ix = st.create_index(s)
        return ix
    
    def _one_segment_index(self):
        ix = self._create_index()
        w = ix.writer()
        w.add_document(f1 = u"A B C", f2 = u"1 2 3", f3 = u"X Y Z")
        w.add_document(f1 = u"D E F", f2 = u"4 5 6", f3 = u"Q R S")
        w.add_document(f1 = u"A E C", f2 = u"1 4 6", f3 = u"X Q S")
        w.add_document(f1 = u"A A A", f2 = u"2 3 5", f3 = u"Y R Z")
        w.add_document(f1 = u"A B", f2 = u"1 2", f3 = u"X Y")
        w.commit()
        
        return ix
    
    def _multi_segment_index(self):
        ix = self._create_index()
        w = ix.writer()
        w.add_document(f1 = u"A B C", f2 = u"1 2 3", f3 = u"X Y Z")
        w.add_document(f1 = u"D E F", f2 = u"4 5 6", f3 = u"Q R S")
        w.commit()
        
        w = ix.writer()
        w.add_document(f1 = u"A E C", f2 = u"1 4 6", f3 = u"X Q S")
        w.add_document(f1 = u"A A A", f2 = u"2 3 5", f3 = u"Y R Z")
        w.commit(NO_MERGE)
        
        w = ix.writer()
        w.add_document(f1 = u"A B", f2 = u"1 2", f3 = u"X Y")
        w.commit(NO_MERGE)
        
        return ix
    
    def test_readers(self):
        target = [(0, u'A', 4, 6), (0, u'B', 2, 2), (0, u'C', 2, 2),
                  (0, u'D', 1, 1), (0, u'E', 2, 2), (0, u'F', 1, 1),
                  (1, u'1', 3, 3), (1, u'2', 3, 3), (1, u'3', 2, 2),
                  (1, u'4', 2, 2), (1, u'5', 2, 2), (1, u'6', 2, 2),
                  (2, u'Q', 2, 2), (2, u'R', 2, 2), (2, u'S', 2, 2),
                  (2, u'X', 3, 3), (2, u'Y', 3, 3), (2, u'Z', 2, 2)]
        
        stored = [{"f1": "A B C"}, {"f1": "D E F"}, {"f1": "A E C"},
                  {"f1": "A A A"}, {"f1": "A B"}]
        
        def t(ix):
            tr = ix.term_reader()
            self.assertEqual(list(tr), target)
            
            dr = ix.doc_reader()
            self.assertEqual(list(dr), stored)
        
        ix = self._one_segment_index()
        self.assertEqual(len(ix.segments), 1)
        t(ix)
        
        ix = self._multi_segment_index()
        self.assertEqual(len(ix.segments), 3)
        t(ix)
    
    def test_term_inspection(self):
        schema = fields.Schema(title=fields.TEXT(stored=True),
                               content=fields.TEXT)
        st = RamStorage()
        ix = st.create_index(schema)
        writer = ix.writer()
        writer.add_document(title=u"My document",
                            content=u"AA AA BB BB CC AA AA AA BB BB CC DD EE EE")
        writer.add_document(title=u"My other document",
                            content=u"AA AB BB CC EE EE AX AX DD")
        writer.commit()
        
        searcher = ix.searcher()
        self.assertEqual(list(searcher.lexicon("content")),
                         [u'aa', u'ab', u'ax', u'bb', u'cc', u'dd', u'ee'])
        self.assertEqual(list(searcher.expand_prefix("content", "a")),
                         [u'aa', u'ab', u'ax'])
        self.assertEqual(list(searcher.all_terms()),
                         [('content', u'aa'), ('content', u'ab'), ('content', u'ax'),
                          ('content', u'bb'), ('content', u'cc'), ('content', u'dd'),
                          ('content', u'ee'), ('title', u'document'), ('title', u'my'),
                          ('title', u'other')])
        # (text, doc_freq, index_freq)
        self.assertEqual(list(searcher.iter_field("content")),
                         [(u'aa', 2, 6), (u'ab', 1, 1), (u'ax', 1, 2),
                          (u'bb', 2, 5), (u'cc', 2, 3), (u'dd', 2, 2),
                          (u'ee', 2, 4)])
        self.assertEqual(list(searcher.iter_field("content", prefix="c")),
                         [(u'cc', 2, 3), (u'dd', 2, 2), (u'ee', 2, 4)])
        self.assertEqual(list(searcher.most_frequent_terms("content")),
                         [(6, u'aa'), (5, u'bb'), (4, u'ee'), (3, u'cc'), (2, u'dd')])
        self.assertEqual(list(searcher.most_frequent_terms("content", prefix="a")),
                         [(6, u'aa'), (2, u'ax'), (1, u'ab')])
    
    def test_vector_postings(self):
        s = fields.Schema(id=fields.ID(stored=True, unique=True),
                          content=fields.TEXT(vector=fields.Positions(analyzer=analysis.StandardAnalyzer())))
        st = RamStorage()
        ix = st.create_index(s)
        
        writer = ix.writer()
        writer.add_document(id=u'1', content=u'the quick brown fox jumped over the lazy dogs')
        writer.commit()
        dr = ix.doc_reader()
        
        terms = list(dr.vector_as(0, 0, "weight"))
        self.assertEqual(terms, [(u'brown', 1.0),
                                 (u'dogs', 1.0),
                                 (u'fox', 1.0),
                                 (u'jumped', 1.0),
                                 (u'lazy', 1.0),
                                 (u'over', 1.0),
                                 (u'quick', 1.0),
                                 ])

        
if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from threading import Lock

from whoosh.util import protected
from whoosh.fields import FieldConfigurationError, UnknownFieldError
from whoosh.reading import DocReader, TermReader, TermNotFound
from whoosh.filedb.filetables import open_docs_table, open_term_table, open_vector_table


# Reader classes

class FileDocReader(DocReader):
    def __init__(self, storage, segment, schema):
        self.storage = storage
        self.segment = segment
        self.schema = schema
        self._scorable_fields = schema.scorable_fields()
        
        from whoosh.filedb.filewriting import DOCLENGTH_TYPE
        recordformat = "<" + DOCLENGTH_TYPE * len(self._scorable_fields)
        self.doclength_table = storage.open_records(segment.doclen_filename, recordformat)
        self.docs_table = open_docs_table(self.storage, segment, schema)
        #self.cache = FifoCache()
        
        self.vector_table = None
        self.is_closed = False
        self._sync_lock = Lock()
        
        self._fieldnum_to_pos = dict((fieldnum, i) for i, fieldnum
                                     in enumerate(schema.scorable_fields()))
    
    def _open_vectors(self):
        if not self.vector_table:
            self.vector_table = open_vector_table(self.storage, self.segment)
    
    @protected
    def __getitem__(self, docnum):
        return self.docs_table.get(docnum)
    
    @protected
    def __iter__(self):
        is_deleted = self.segment.is_deleted
        for docnum in xrange(0, self.segment.max_doc):
            if not is_deleted(docnum):
                yield self.docs_table.get(docnum)
    
    def close(self):
        self.doclength_table.close()
        self.docs_table.close()
        if self.vector_table:
            self.vector_table.close()
        self.is_closed = True
    
    def doc_count_all(self):
        return self.segment.doc_count_all()
    
    def doc_count(self):
        return self.segment.doc_count()
    
    def field_length(self, fieldid):
        fieldid = self.schema.to_number(fieldid)
        return self.segment.field_length(fieldid)
    
    @protected
    def doc_field_length(self, docnum, fieldid):
        fieldid = self.schema.to_number(fieldid)
        if fieldid not in self._scorable_fields:
            raise FieldConfigurationError("Field %r does not store lengths" % fieldid)
        
        pos = self._fieldnum_to_pos[fieldid]
        return self.doclength_table.get(docnum, pos)
    
    @protected
    def doc_field_lengths(self, docnum):
        return self.doclength_table.get_record(docnum)
    
    @protected
    def vector(self, docnum, fieldid):
        """Yields a sequence of raw (text, data) tuples representing
        the term vector for the given document and field.
        """
        
        self._open_vectors()
        fieldnum = self.schema.to_number(fieldid)
        readfn = self.vector_format(fieldnum).read_postvalue
        return self.vector_table.postings((docnum, fieldnum), readfn)
    
    def vector_as(self, docnum, fieldnum, astype):
        format = self.vector_format(fieldnum)
        
        if format is None:
            raise FieldConfigurationError("Field %r is not vectored" % self.schema.number_to_name(fieldnum))
        elif not format.supports(astype):
            raise FieldConfigurationError("Field %r does not support %r" % (self.schema.number_to_name(fieldnum),
                                                                            astype))
        
        interpreter = format.interpreter(astype)
        for text, data in self.vector(docnum, fieldnum):
            yield (text, interpreter(data))
    

class FileTermReader(TermReader):
    """
    Do not instantiate this object directly. Instead use Index.term_reader().
    
    Reads term information from a segment.
    
    Each TermReader represents two open files. Remember to close() the reader when
    you're done with it.
    """
    
    def __init__(self, storage, segment, schema):
        """
        :param storage: The storage object in which the segment resides.
        :param segment: The segment to read from.
        :param schema: The index's schema object.
        """
        
        self.segment = segment
        self.schema = schema
        
        self.term_table = open_term_table(storage, segment)
        self.is_closed = False
        self._sync_lock = Lock()
    
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.segment)
    
    @protected
    def __iter__(self):
        tt = self.term_table
        for (fn, t), termcount in tt:
            yield (fn, t, tt.posting_count((fn, t)), termcount)
    
    @protected
    def __contains__(self, term):
        return (self.schema.to_number(term[0]), term[1]) in self.term_table
    
    def close(self):
        self.term_table.close()
        self.is_closed = True
    
    def format(self, fieldname):
        if fieldname in self.schema:
            return self.schema.field_by_name(fieldname).format
        else:
            raise UnknownFieldError(fieldname)
    
    @protected
    def _term_info(self, fieldnum, text):
        try:
            return self.term_table.get((fieldnum, text))
        except KeyError:
            raise TermNotFound("%s:%r" % (fieldnum, text))
    
    @protected
    def doc_frequency(self, fieldid, text):
        fieldid = self.schema.to_number(fieldid)
        if (fieldid, text) not in self.term_table:
            return 0
        return self.term_table.posting_count((fieldid, text))
    
    @protected
    def frequency(self, fieldid, text):
        fieldid = self.schema.to_number(fieldid)
        if (fieldid, text) not in self.term_table:
            return 0
        return self.term_table.get((fieldid, text))
    
    def doc_count_all(self):
        return self.segment.doc_count_all()
    
    @protected
    def iter_from(self, fieldnum, text):
        tt = self.term_table
        postingcount = tt.posting_count
        for (fn, t), termcount in tt.items_from((fieldnum, text)):
            yield (fn, t, postingcount((fn, t)), termcount)
    
    @protected
    def lexicon(self, fieldid):
        # The base class has a lexicon() implementation that uses
        # iter_from() and throws away the value, but overriding to
        # use FileTableReader.keys_from() is much, much faster.
        
        tt = self.term_table
        fieldid = self.schema.to_number(fieldid)
        for fn, t in tt.keys_from((fieldid, '')):
            if fn != fieldid:
                return
            yield t
    
    @protected
    def expand_prefix(self, fieldid, prefix):
        # The base class has an expand_prefix() implementation that uses
        # iter_from() and throws away the value, but overriding to
        # use FileTableReader.keys_from() is much, much faster.
        
        tt = self.term_table
        fieldid = self.schema.to_number(fieldid)
        for fn, t in tt.keys_from((fieldid, prefix)):
            if fn != fieldid or not t.startswith(prefix):
                return
            yield t
    
    # Posting retrieval methods
    
    @protected
    def postings(self, fieldnum, text, exclude_docs = None):
        is_deleted = self.segment.is_deleted
        no_exclude = exclude_docs is None
        
        # The format object is actually responsible for parsing the
        # posting data from disk.
        readfn = self.schema.field_by_number(fieldnum).format.read_postvalue
        
        for docnum, data in self.term_table.postings((fieldnum, text), readfn = readfn):
            if not is_deleted(docnum)\
               and (no_exclude or docnum not in exclude_docs):
                yield docnum, data
    
    def postings_as(self, fieldid, text, astype, exclude_docs = None):
        format = self.schema[fieldid].format
        if not format.supports(astype):
            raise FieldConfigurationError("Field %s format does not support %r" % (self.schema.to_name(fieldid),
                                                                                   astype))
        
        interp = format.interpreter(astype)
        for docnum, data in self.postings(fieldid, text, exclude_docs = exclude_docs):
            yield (docnum, interp(data))
    
















    
    
    


import unittest

from whoosh import fields, index, qparser, query, searching, scoring
from whoosh.filedb.filestore import RamStorage
from whoosh.query import *

class TestReading(unittest.TestCase):
    def make_index(self):
        s = fields.Schema(key = fields.ID(stored = True),
                          name = fields.TEXT,
                          value = fields.TEXT)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(key = u"A", name = u"Yellow brown", value = u"Blue red green render purple?")
        w.add_document(key = u"B", name = u"Alpha beta", value = u"Gamma delta epsilon omega.")
        w.add_document(key = u"C", name = u"One two", value = u"Three rendered four five.")
        w.add_document(key = u"D", name = u"Quick went", value = u"Every red town.")
        w.add_document(key = u"E", name = u"Yellow uptown", value = u"Interest rendering outer photo!")
        w.commit()
        
        return ix
    
    def _get_keys(self, stored_fields):
        return sorted([d.get("key") for d in stored_fields])
    
    def _docs(self, q, s):
        return self._get_keys([s.stored_fields(docnum) for docnum
                               in q.docs(s)])
    
    def _doc_scores(self, q, s, w):
        return self._get_keys([s.stored_fields(docnum) for docnum, score
                               in q.doc_scores(s, weighting = w)])
    
    def test_empty_index(self):
        schema = fields.Schema(key = fields.ID(stored=True), value = fields.TEXT)
        st = RamStorage()
        self.assertRaises(index.EmptyIndexError, st.open_index, schema)
    
    def test_docs_method(self):
        ix = self.make_index()
        s = ix.searcher()
        
        self.assertEqual(self._get_keys(s.documents(name = "yellow")), [u"A", u"E"])
        self.assertEqual(self._get_keys(s.documents(value = "red")), [u"A", u"D"])
    
    def test_queries(self):
        ix = self.make_index()
        s = ix.searcher()
        
        tests = [
                 (Term("name", u"yellow"),
                  [u"A", u"E"]),
                 (Term("value", u"red"),
                  [u"A", u"D"]),
                 (Term("value", u"zeta"),
                  []),
                 (Require(Term("value", u"red"), Term("name", u"yellow")),
                  [u"A"]),
                 (And([Term("value", u"red"), Term("name", u"yellow")]),
                  [u"A"]),
                 (Or([Term("value", u"red"), Term("name", u"yellow")]),
                  [u"A", u"D", u"E"]),
                 (Or([Term("value", u"red"), Term("name", u"yellow"), Not(Term("name", u"quick"))]),
                  [u"A", u"E"]),
                 (AndNot(Term("name", u"yellow"), Term("value", u"purple")),
                  [u"E"]),
                 (Variations("value", u"render"), [u"A", u"C", u"E"]),
                 (Or([Wildcard('value', u'*red*'), Wildcard('name', u'*yellow*')]),
                  [u"A", u"C", u"D", u"E"]),
                ]
        
        for query, result in tests:
            self.assertEqual(self._docs(query, s), result)
        
        for wcls in dir(scoring):
            if wcls is scoring.Weighting: continue
            if isinstance(wcls, scoring.Weighting):
                for query, result in tests:
                    self.assertEqual(self._doc_scores(query, s, wcls), result)
        
        for methodname in ("_docs", "_doc_scores"):
            method = getattr(self, methodname)

    def test_keyword_or(self):
        schema = fields.Schema(a=fields.ID(stored=True), b=fields.KEYWORD)
        st = RamStorage()
        ix = st.create_index(schema)
        
        w = ix.writer()
        w.add_document(a=u"First", b=u"ccc ddd")
        w.add_document(a=u"Second", b=u"aaa ddd")
        w.add_document(a=u"Third", b=u"ccc eee")
        w.commit()
        
        qp = qparser.QueryParser("b", schema=schema)
        searcher = ix.searcher()
        qr = qp.parse("b:ccc OR b:eee")
        self.assertEqual(qr.__class__, query.Or)
        r = searcher.search(qr)
        self.assertEqual(len(r), 2)
        self.assertEqual(r[0]["a"], "Third")
        self.assertEqual(r[1]["a"], "First")

    def test_score_retrieval(self):
        schema = fields.Schema(title=fields.TEXT(stored=True),
                               content=fields.TEXT(stored=True))
        storage = RamStorage()
        ix = storage.create_index(schema)
        writer = ix.writer()
        writer.add_document(title=u"Miss Mary",
                            content=u"Mary had a little white lamb its fleece was white as snow")
        writer.add_document(title=u"Snow White",
                            content=u"Snow white lived in the forrest with seven dwarfs")
        writer.commit()
        
        searcher = ix.searcher()
        results = searcher.search(Term("content", "white"))
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0]['title'], u"Miss Mary")
        self.assertEqual(results[1]['title'], u"Snow White")
        self.assertNotEqual(results.score(0), None)
        self.assertNotEqual(results.score(0), 0)
        self.assertNotEqual(results.score(0), 1)

    def test_missing_field_scoring(self):
        schema = fields.Schema(name=fields.TEXT(stored=True), hobbies=fields.TEXT(stored=True))
        storage = RamStorage()
        idx = storage.create_index(schema)
        writer = idx.writer() 
        writer.add_document(name=u'Frank', hobbies=u'baseball, basketball')
        writer.commit()
        self.assertEqual(idx.segments[0].field_length(0), 2) # hobbies
        self.assertEqual(idx.segments[0].field_length(1), 1) # name
        
        writer = idx.writer()
        writer.add_document(name=u'Jonny') 
        writer.commit()
        self.assertEqual(len(idx.segments), 1)
        self.assertEqual(idx.segments[0].field_length(0), 2) # hobbies
        self.assertEqual(idx.segments[0].field_length(1), 2) # name
        
        parser = qparser.MultifieldParser(['name', 'hobbies'], schema=schema)
        searcher = idx.searcher()
        result = searcher.search(parser.parse(u'baseball'))
        self.assertEqual(len(result), 1)


if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from array import array
from collections import defaultdict

from whoosh.fields import UnknownFieldError
from whoosh.writing import IndexWriter, IndexingError
from whoosh.filedb import postpool
from whoosh.filedb.fileindex import SegmentDeletionMixin, Segment, SegmentSet
from whoosh.filedb.filereading import FileTermReader, FileDocReader
from whoosh.filedb.filetables import copy_postings
from whoosh.filedb.filetables import create_docs_table, create_term_table, create_vector_table
from whoosh.util import fib


DOCLENGTH_TYPE = "H"
DOCLENGTH_LIMIT = 2**16-1


# Merge policies

# A merge policy is a callable that takes the Index object,
# the SegmentWriter object, and the current SegmentSet
# (not including the segment being written), and returns an
# updated SegmentSet (not including the segment being
# written).

def NO_MERGE(ix, writer, segments):
    """This policy does not merge any existing segments.
    """
    return segments


def MERGE_SMALL(ix, writer, segments):
    """This policy merges small segments, where small is
    defined using a heuristic based on the fibonacci sequence.
    """
    
    newsegments = SegmentSet()
    sorted_segment_list = sorted((s.doc_count_all(), s) for s in segments)
    total_docs = 0
    for i, (count, seg) in enumerate(sorted_segment_list):
        if count > 0:
            total_docs += count
            if total_docs < fib(i + 5):
                writer.add_segment(ix, seg)
            else:
                newsegments.append(seg)
    return newsegments


def OPTIMIZE(ix, writer, segments):
    """This policy merges all existing segments.
    """
    for seg in segments:
        writer.add_segment(ix, seg)
    return SegmentSet()


# Writing classes

class FileIndexWriter(SegmentDeletionMixin, IndexWriter):
    # This class is mostly a shell for SegmentWriter. It exists to handle
    # multiple SegmentWriters during merging/optimizing.
    
    def __init__(self, ix, postlimit = 4 * 1024 * 1024):
        """
        :param ix: the Index object you want to write to.
        :param postlimit: Essentially controls the maximum amount of memory the
            indexer uses at a time, in bytes (the actual amount of memory used by
            the Python process will be much larger because of other overhead).
            The default (4MB) is quite small. You should increase this value
            for large collections, e.g. ``postlimit=32*1024*1024``.
        """
        
        # Obtain a lock
        self.locked = ix.lock()
        
        self.index = ix
        self.segments = ix.segments.copy()
        self.postlimit = postlimit
        self._segment_writer = None
        self._searcher = ix.searcher()
    
    def _finish(self):
        self._close_searcher()
        self._segment_writer = None
        # Release the lock
        if self.locked:
            self.index.unlock()
    
    def segment_writer(self):
        """Returns the underlying SegmentWriter object."""
        
        if not self._segment_writer:
            self._segment_writer = SegmentWriter(self.index, self.postlimit)
        return self._segment_writer
    
    def add_document(self, **fields):
        self.segment_writer().add_document(fields)
    
    def commit(self, mergetype = MERGE_SMALL):
        """Finishes writing and unlocks the index.
        
        :param mergetype: How to merge existing segments. One of
            :class:`whoosh.filedb.filewriting.NO_MERGE`,
            :class:`whoosh.filedb.filewriting.MERGE_SMALL`,
            or :class:`whoosh.filedb.filewriting.OPTIMIZE`.
        """
        
        self._close_searcher()
        if self._segment_writer or mergetype is OPTIMIZE:
            self._merge_segments(mergetype)
        self.index.commit(self.segments)
        self._finish()
        
    def cancel(self):
        self._finish()
    
    def _merge_segments(self, mergetype):
        sw = self.segment_writer()
        new_segments = mergetype(self.index, sw, self.segments)
        sw.close()
        new_segments.append(sw.segment())
        self.segments = new_segments


class SegmentWriter(object):
    """
    Do not instantiate this object directly; it is created by the IndexWriter object.
    
    Handles the actual writing of new documents to the index: writes stored fields,
    handles the posting pool, and writes out the term index.
    """
    
    def __init__(self, ix, postlimit, name = None):
        """
        :param ix: the Index object in which to write the new segment.
        :param postlimit: the maximum size for a run in the posting pool.
        :param name: the name of the segment.
        """
        
        self.index = ix
        self.schema = ix.schema
        self.storage = ix.storage
        self.name = name or ix._next_segment_name()
        
        self.max_doc = 0

        self.pool = postpool.PostingPool(limit = postlimit)
        self._scorable_to_pos = dict((fnum, i) for i, fnum in enumerate(self.schema.scorable_fields()))
        self._stored_to_pos = dict((fnum, i) for i, fnum in enumerate(self.schema.stored_fields()))
        
        # Create a temporary segment object just so we can access
        # its *_filename attributes (so if we want to change the
        # naming convention, we only have to do it in one place).
        tempseg = Segment(self.name, 0, 0, None)
        
        # Open files for writing
        self.term_table = create_term_table(self.storage, tempseg)
        self.docs_table = create_docs_table(self.storage, tempseg)
        
        recordformat = "<" + DOCLENGTH_TYPE * len(self._scorable_to_pos)
        self.doclength_table = self.storage.create_records(tempseg.doclen_filename,
                                                           recordformat)
        
        self.vector_table = None
        if self.schema.has_vectored_fields():
            self.vector_table = create_vector_table(self.storage, tempseg)
        
        # Keep track of the total number of tokens (across all docs)
        # in each field
        self.field_length_totals = defaultdict(int)
            
    def segment(self):
        """Returns an index.Segment object for the segment being written."""
        return Segment(self.name, self.max_doc, dict(self.field_length_totals))
    
    def close(self):
        """Finishes writing the segment (flushes the posting pool out to disk) and
        closes all open files.
        """
        
        self._flush_pool()
        
        self.doclength_table.close()
        
        self.docs_table.close()
        self.term_table.close()
        
        if self.vector_table:
            self.vector_table.close()
        
    def add_index(self, other_ix):
        """Adds the contents of another Index object to this segment.
        This currently does NO checking of whether the schemas match up.
        """
        
        for seg in other_ix.segments:
            self.add_segment(other_ix, seg)

    def add_segment(self, ix, segment):
        """Adds the contents of another segment to this one. This is used
        to merge existing segments into the new one before deleting them.
        
        :param ix: The index.Index object containing the segment to merge.
        :param segment: The index.Segment object to merge into this one.
        """
        
        start_doc = self.max_doc
        has_deletions = segment.has_deletions()
        
        if has_deletions:
            doc_map = {}
        
        # Merge document info
        docnum = 0
        schema = ix.schema
        name2num = schema.name_to_number
        stored_to_pos = self._stored_to_pos
        
        def storedkeyhelper(item):
            return stored_to_pos[name2num(item[0])]
        
        doc_reader = FileDocReader(ix.storage, segment, schema)
        try:
            vectored_fieldnums = ix.schema.vectored_fields()
            if vectored_fieldnums:
                doc_reader._open_vectors()
                inv = doc_reader.vector_table
                outv = self.vector_table
            
            for docnum in xrange(segment.max_doc):
                if not segment.is_deleted(docnum):
                    # Copy the stored fields and field lengths from the other segment into this one
                    storeditems = doc_reader[docnum].items()
                    storedvalues = [v for k, v in sorted(storeditems, key=storedkeyhelper)]
                    self._add_doc_data(storedvalues, doc_reader.doc_field_lengths(docnum))
                    
                    if has_deletions:
                        doc_map[docnum] = self.max_doc
                    
                    # Copy term vectors
                    for fieldnum in vectored_fieldnums:
                        if (docnum, fieldnum) in inv:
                            copy_postings(inv, (docnum, fieldnum),
                                          outv, (self.max_doc, fieldnum))
                
                    self.max_doc += 1
            
            # Add field length totals
            for fieldnum, total in segment.field_length_totals.iteritems():
                self.field_length_totals[fieldnum] += total
        
        finally:
            doc_reader.close()
        
        # Merge terms
        term_reader = FileTermReader(ix.storage, segment, ix.schema)
        try:
            for fieldnum, text, _, _ in term_reader:
                for docnum, data in term_reader.postings(fieldnum, text):
                    if has_deletions:
                        newdoc = doc_map[docnum]
                    else:
                        newdoc = start_doc + docnum
                    
                    self.pool.add_posting(fieldnum, text, newdoc, data)
        finally:
            term_reader.close()

    def add_document(self, fields):
        scorable_to_pos = self._scorable_to_pos
        stored_to_pos = self._stored_to_pos
        
        fieldnames = [name for name in fields.keys() if not name.startswith("_")]
        
        schema = self.schema
        for name in fieldnames:
            if name not in schema:
                raise UnknownFieldError("There is no field named %r" % name)
        
        fieldlengths = array(DOCLENGTH_TYPE, [0] * len(scorable_to_pos))
        storedvalues = [None] * len(stored_to_pos)
        
        fieldnames.sort(key = schema.name_to_number)
        for name in fieldnames:
            value = fields.get(name)
            
            if value:
                fieldnum = schema.name_to_number(name)
                field = schema.field_by_name(name)
                format = field.format
                
                # If the field is indexed, add the words in the value to the index
                if format.analyzer:
                    if not isinstance(value, unicode):
                        raise ValueError("%r in field %s is not unicode" % (value, name))
                    
                    # Count of all terms in the value
                    count = 0
                    # Count of UNIQUE terms in the value
                    unique = 0
                    
                    # TODO: Method for adding progressive field values, ie setting
                    # start_pos/start_char?
                    for w, freq, data in format.word_datas(value):
                        #assert w != ""
                        self.pool.add_posting(fieldnum, w, self.max_doc, data)
                        count += freq
                        unique += 1
                    
                    if field.scorable:
                        # Add the term count to the total for this field
                        self.field_length_totals[fieldnum] += count
                        # Set the term count to the per-document field length
                        fieldlengths[scorable_to_pos[fieldnum]] = count
                
                # If the field is vectored, add the words in the value to
                # the vector table
                vector = field.vector
                if vector:
                    vtable = self.vector_table
                    # TODO: Method for adding progressive field values, ie setting
                    # start_pos/start_char?
                    vdata = dict((w, data) for w, freq, data in vector.word_datas(value,))
                    write_postvalue = vector.write_postvalue
                    for word in sorted(vdata.keys()):
                        vtable.write_posting(word, vdata[word], writefn = write_postvalue)
                    vtable.add((self.max_doc, fieldnum), '')
                
                # If the field is stored, add the value to the doc state
                if field.stored:
                    storedname = "_stored_" + name
                    if storedname in fields:
                        stored_value = fields[storedname]
                    else :
                        stored_value = value
                    storedvalues[stored_to_pos[fieldnum]] = stored_value
        
        self._add_doc_data(storedvalues, fieldlengths)
        self.max_doc += 1
    
    def _add_doc_data(self, storedvalues, fieldlengths):
        self.docs_table.append(storedvalues)
        self.doclength_table.append(fieldlengths)
    
    def _flush_pool(self):
        # This method pulls postings out of the posting pool (built up
        # as documents are added) and writes them to the posting file.
        # Each time it encounters a posting for a new term, it writes
        # the previous term to the term index (by waiting to write the
        # term entry, we can easily count the document frequency and
        # sum the terms by looking at the postings).
        
        term_table = self.term_table
        
        write_posting_method = None
        current_fieldnum = None # Field number of the current term
        current_text = None # Text of the current term
        first = True
        current_weight = 0
        
        # Loop through the postings in the pool.
        # Postings always come out of the pool in field number/alphabetic order.
        for fieldnum, text, docnum, data in self.pool:
            # If we're starting a new term, reset everything
            if write_posting_method is None or fieldnum > current_fieldnum or text > current_text:
                if fieldnum != current_fieldnum:
                    write_posting_method = self.schema.field_by_number(fieldnum).format.write_postvalue
                    
                
                # If we've already written at least one posting, write the
                # previous term to the index.
                if not first:
                    term_table.add((current_fieldnum, current_text), current_weight)
                    
                # Reset term variables
                current_fieldnum = fieldnum
                current_text = text
                current_weight = 0
                first = False
            
            elif fieldnum < current_fieldnum or (fieldnum == current_fieldnum and text < current_text):
                # This should never happen!
                raise Exception("Postings are out of order: %s:%s .. %s:%s" %
                                (current_fieldnum, current_text, fieldnum, text))
            
            current_weight += term_table.write_posting(docnum, data, write_posting_method)
        
        # Finish up the last term
        if not first:
            term_table.add((current_fieldnum, current_text), current_weight)




        
        



#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""
This module contains classes that allow reading from an index.
"""

from bisect import bisect_right
from heapq import heapify, heapreplace, heappop, nlargest
from threading import Lock

from whoosh.fields import UnknownFieldError
from whoosh.util import ClosableMixin

# Exceptions

class TermNotFound(Exception):
    pass


# Base classes

class DocReader(ClosableMixin):
    """Do not instantiate this object directly. Instead use Index.doc_reader().
    """

    def __getitem__(self, docnum):
        """Returns the stored fields for the given document number.
        """
        raise NotImplementedError
    
    def __iter__(self):
        """Yields the stored fields for all documents.
        """
        raise NotImplementedError
    
    def close(self):
        """Closes the open files associated with this reader.
        """
        raise NotImplementedError
    
    def doc_count_all(self):
        """Returns the total number of documents, DELETED OR UNDELETED,
        in this reader.
        """
        raise NotImplementedError
    
    def doc_count(self):
        """Returns the total number of UNDELETED documents in this reader.
        """
        raise NotImplementedError
    
    def field_length(self, fieldid):
        """Returns the total number of terms in the given field. This is used
        by some scoring algorithms.
        """
        raise NotImplementedError
    
    def doc_field_length(self, docnum, fieldid):
        """Returns the number of terms in the given field in the
        given document. This is used by some scoring algorithms.
        """
        raise NotImplementedError
    
    def doc_field_lengths(self, docnum):
        """Returns an array corresponding to the lengths of the
        scorable fields in the given document. It's up to the
        caller to correlate the positions of the numbers in the
        array with the scorable fields in the schema.
        """
        raise NotImplementedError
    
    def vector_format(self, fieldid):
        """Returns the vector format object associated with the given
        field, or None if the field does not store term vectors.
        """
        return self.schema[fieldid].vector
    
    def vector_supports(self, fieldid, astype):
        """Returns True if the vector format for the given field supports
        the given data interpretation.
        
        :param astype: a string containing the name of the format you
            want to check the vector supports, for example "weights".
        """
        format = self.vector_format(fieldid)
        if format is None: return False
        return format.supports(astype)
    
    def vector(self, docnum, fieldid):
        raise NotImplementedError
    
    def vector_as(self, docnum, fieldid, astype):
        """Yields a sequence of interpreted (text, data) tuples
        representing the term vector for the given document and
        field.
        
        :param astype: a string containing the name of the format you
            want the term vector's data in, for example "weights".
        """
        raise NotImplementedError
    

class TermReader(ClosableMixin):
    def __iter__(self):
        """Yields (fieldnum, token, docfreq, indexfreq) tuples for
        each term in the reader, in lexical order.
        """
        raise NotImplementedError
    
    def __contains__(self, term):
        """Returns True if the given term tuple (fieldid, text) is
        in this reader.
        """
        raise NotImplementedError
    
    def close(self):
        """Closes the open files associated with this reader.
        """
        raise NotImplementedError
    
    def format(self, fieldid):
        """Returns the Format object corresponding to the given field name.
        """
        if fieldid in self.schema:
            return self.schema[fieldid].format
        else:
            raise UnknownFieldError(fieldid)
    
    def doc_frequency(self, fieldid, text):
        """Returns the document frequency of the given term (that is,
        how many documents the term appears in).
        """
        raise NotImplementedError
    
    def frequency(self, fieldid, text):
        """Returns the total number of instances of the given term
        in the collection.
        """
        raise NotImplementedError
    
    def doc_count_all(self):
        """Returns the total number of documents, DELETED OR UNDELETED,
        in this reader.
        """
        raise NotImplementedError
    
    def iter_from(self, fieldnum, text):
        """Yields (field_num, text, doc_freq, index_freq) tuples
        for all terms in the reader, starting at the given term.
        """
        raise NotImplementedError
    
    def expand_prefix(self, fieldid, prefix):
        """Yields terms in the given field that start with the given prefix.
        """
        raise NotImplementedError
    
    def all_terms(self):
        """Yields (fieldname, text) tuples for every term in the index.
        """
        
        num2name = self.schema.number_to_name
        current_fieldnum = None
        current_fieldname = None
        
        for fn, t, _, _ in self:
            # Only call self.schema.number_to_name when the
            # field number changes.
            if fn != current_fieldnum:
                current_fieldnum = fn
                current_fieldname = num2name(fn)
            yield (current_fieldname, t)
    
    def iter_field(self, fieldid, prefix = ''):
        """Yields (text, doc_freq, index_freq) tuples for all terms
        in the given field.
        """
        
        fieldid = self.schema.to_number(fieldid)
        for fn, t, docfreq, freq in self.iter_from(fieldid, prefix):
            if fn != fieldid:
                return
            yield t, docfreq, freq
    
    def iter_prefix(self, fieldid, prefix):
        """Yields (field_num, text, doc_freq, index_freq) tuples
        for all terms in the given field with a certain prefix.
        """
        
        fieldid = self.schema.to_number(fieldid)
        for fn, t, docfreq, colfreq in self.iter_from(fieldid, prefix):
            if fn != fieldid or not t.startswith(prefix):
                return
            yield (t, docfreq, colfreq)
    
    def most_frequent_terms(self, fieldid, number = 5, prefix = None):
        """Yields the top 'number' most frequent terms in the given field as
        a series of (frequency, text) tuples.
        """
        
        if prefix is not None:
            iterator = self.iter_prefix(fieldid, prefix)
        else:
            iterator = self.iter_field(fieldid)
        
        return nlargest(number,
                        ((indexfreq, token)
                         for token, _, indexfreq
                         in iterator))
        
    def lexicon(self, fieldid):
        """Yields all terms in the given field."""
        
        for t, _, _ in self.iter_field(fieldid):
            yield t
    
    def postings(self, fieldid, text):
        raise NotImplementedError
    
    def postings_as(self, fieldid, text, astype, exclude_docs = None):
        """Yields interpreted data for each document containing
        the given term. The current field must have stored positions
        for this to work.
        
        :param astype:
            how to interpret the posting data, for example
            "positions". The field must support the interpretation.
        :param exclude_docs:
            a set of document numbers to ignore. This
            is used by queries to skip documents that have already been
            eliminated from consideration.
        :param boost: a factor by which to multiply each weight.
        """
        raise NotImplementedError
    
    def weights(self, fieldid, text, exclude_docs = None):
        """
        Yields (docnum, term_weight) tuples for each document containing
        the given term. The current field must have stored term weights
        for this to work.
        
        :param exclude_docs:
            a set of document numbers to ignore. This
            is used by queries to skip documents that have already been
            eliminated from consideration.
        """
        
        return self.postings_as(fieldid, text, "weight", exclude_docs = exclude_docs)
    
    def positions(self, fieldid, text, exclude_docs = None):
        """Yields (docnum, [positions]) tuples for each document containing
        the given term. The current field must have stored positions
        for this to work.
        
        :param exclude_docs:
            a set of document numbers to ignore. This
            is used by queries to skip documents that have already been
            eliminated from consideration.
        :param boost: a factor by which to multiply each weight.
        """
        
        return self.postings_as(fieldid, text, "positions", exclude_docs = exclude_docs)


# Multisegment reader classes

class MultiDocReader(DocReader):
    """
    Do not instantiate this object directly. Instead use Index.doc_reader().
    
    Reads document-related information by aggregating the results from
    multiple segments. The main interface is to either iterate on this
    object to yield the document stored fields, or use getitem (e.g. docreader[10])
    to get the stored fields for a specific document number.
    
    Each MultiDocReader represents (number of segments * 2) open files.
    Be sure to close() the reader when you're finished with it.
    """
    
    def __init__(self, doc_readers, doc_offsets, schema):
        self.doc_readers = doc_readers
        self.doc_offsets = doc_offsets
        self.schema = schema
        self._scorable_fields = self.schema.scorable_fields()
        
        self.is_closed = False
        self._sync_lock = Lock()
        
    def __getitem__(self, docnum):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.doc_readers[segmentnum].__getitem__(segmentdoc)
    
    def __iter__(self):
        for reader in self.doc_readers:
            for result in reader:
                yield result
    
    def close(self):
        """Closes the open files associated with this reader.
        """
        
        for d in self.doc_readers:
            d.close()
        self.is_closed = True
    
    def doc_count_all(self):
        return sum(dr.doc_count_all() for dr in self.doc_readers)
    
    def doc_count(self):
        return sum(dr.doc_count() for dr in self.doc_readers)
    
    def field_length(self, fieldnum):
        return sum(dr.field_length(fieldnum) for dr in self.doc_readers)
    
    def doc_field_length(self, docnum, fieldid):
        fieldid = self.schema.to_number(fieldid)
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.doc_readers[segmentnum].doc_field_length(segmentdoc, fieldid)
    
    def doc_field_lengths(self, docnum):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.doc_readers[segmentnum].doc_field_lengths(segmentdoc)
    
    def unique_count(self, docnum):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.doc_readers[segmentnum].unique_count(segmentdoc)
    
    def _document_segment(self, docnum):
        return max(0, bisect_right(self.doc_offsets, docnum) - 1)
    
    def _segment_and_docnum(self, docnum):
        segmentnum = self._document_segment(docnum)
        offset = self.doc_offsets[segmentnum]
        return segmentnum, docnum - offset
    
    def vector(self, docnum, fieldid):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.doc_readers[segmentnum].vector(segmentdoc, fieldid)
    
    def _doc_info(self, docnum, key):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.doc_readers[segmentnum]._doc_info(segmentdoc, key)
    

class MultiTermReader(TermReader):
    """Do not instantiate this object directly. Instead use Index.term_reader().
    
    Reads term information by aggregating the results from
    multiple segments.
    
    Each MultiTermReader represents (number of segments * 2) open files.
    Be sure to close() the reader when you're finished with it.
    """
    
    def __init__(self, term_readers, doc_offsets, schema):
        self.term_readers = term_readers
        self.doc_offsets = doc_offsets
        self.schema = schema
        
        self.is_closed = False
        self._sync_lock = Lock()
    
    def __contains__(self, term):
        return any(tr.__contains__(term) for tr in self.term_readers)
    
    def __iter__(self):
        return self._merge_iters([iter(r) for r in self.term_readers])
    
    def iter_from(self, fieldnum, text):
        return self._merge_iters([r.iter_from(fieldnum, text) for r in self.term_readers])
    
    def close(self):
        for tr in self.term_readers:
            tr.close()
        self.is_closed = True
    
    def doc_frequency(self, fieldnum, text):
        if (fieldnum, text) not in self:
            return 0
        
        return sum(r.doc_frequency(fieldnum, text) for r in self.term_readers)
    
    def frequency(self, fieldnum, text):
        if (fieldnum, text) not in self:
            return 0
        
        return sum(r.frequency(fieldnum, text) for r in self.term_readers)
    
    def _merge_iters(self, iterlist):
        # Merge-sorts terms coming from a list of
        # term iterators (TermReader.__iter__() or
        # TermReader.iter_from()).
        
        # Fill in the list with the head term from each iterator.
        # infos is a list of [headterm, iterator] lists.
        
        current = []
        for it in iterlist:
            fnum, text, docfreq, termcount = it.next()
            current.append((fnum, text, docfreq, termcount, it))
        heapify(current)
        
        # Number of active iterators
        active = len(current)
        while active > 0:
            # Peek at the first term in the sorted list
            fnum, text = current[0][:2]
            docfreq = 0
            termcount = 0
            
            # Add together all terms matching the first
            # term in the list.
            while current and current[0][0] == fnum and current[0][1] == text:
                docfreq += current[0][2]
                termcount += current[0][3]
                it = current[0][4]
                try:
                    fn, t, df, tc = it.next()
                    heapreplace(current, (fn, t, df, tc, it))
                except StopIteration:
                    heappop(current)
                    active -= 1
                
            # Yield the term with the summed frequency and
            # term count.
            yield (fnum, text, docfreq, termcount)
    
    def postings(self, fieldnum, text, exclude_docs = None):
        for i, r in enumerate(self.term_readers):
            offset = self.doc_offsets[i]
            if (fieldnum, text) in r:
                for docnum, data in r.postings(fieldnum, text, exclude_docs = exclude_docs):
                    yield (docnum + offset, data)
                    



if __name__ == '__main__':
    pass












    
    
    


#===============================================================================
# Copyright 2008 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from __future__ import division
from heapq import nlargest

"""
The highlight module contains classes and functions for displaying short
excerpts from hit documents in the search results you present to the user, with
query terms highlighted.
"""

# Fragment object

class Fragment(object):
    """Represents a fragment (extract) from a hit document. This object is mainly
    used to keep track of the start and end points of the fragment; it does not
    contain the text of the fragment or do much else.
    """
    
    def __init__(self, tokens, charsbefore = 0, charsafter = 0, textlen = 999999):
        """
        :param tokens: list of the Token objects representing the matched terms. 
        :param charsbefore: approx. how many characters before the start of the first
            matched term to include in the fragment.
        :param charsafter: approx. how many characters after the end of the last
            matched term to include in the fragment.
        :param textlen: length in characters of the document text.
        """
        
        #: index of the first character of the fragment in the original document
        self.startchar = max(0, tokens[0].startchar - charsbefore)
        #: index after the last character of the fragment in the original document
        self.endchar = min(textlen, tokens[-1].endchar + charsafter)
        self.matches = [t for t in tokens if t.matched]
        self.matched_terms = frozenset(t.text for t in self.matches)
    
    def __len__(self):
        return self.endchar - self.startchar
    
    def overlaps(self, fragment):
        sc = self.startchar
        ec = self.endchar
        fsc = fragment.startchar
        fec = fragment.endchar
        return (fsc > sc and fsc < ec) or (fec > sc and fec < ec)
    
    def overlapped_length(self, fragment):
        sc = self.startchar
        ec = self.endchar
        fsc = fragment.startchar
        fec = fragment.endchar
        return max(ec, fec) - min(sc, fsc)
    
    def has_matches(self):
        return any(t.matched for t in self.tokens)
    

# Filters

def copyandmatchfilter(termset, tokens):
    for t in tokens:
        t = t.copy()
        t.matched = t.text in termset
        yield t

# Fragmenters

def NullFragmenter(text, termset, tokens):
    """Doesn't fragment the token stream. This object just
    returns the entire stream as one "fragment". This is useful if
    you want to highlight the entire text.
    """
    return [Fragment(list(tokens))]


class SimpleFragmenter(object):
    """Simply splits the text into roughly equal sized chunks.
    """
    
    def __init__(self, size = 70):
        """
        :param size: size (in characters) to chunk to. The chunking is based on
            tokens, so the fragments will usually be smaller.
        """
        self.size = 70
        
    def __call__(self, text, tokens):
        size = self.size
        first = None
        frag = []
        
        for t in tokens:
            if first is None:
                first = t.startchar
            
            if t.endchar - first > size:
                first = None
                if frag:
                    yield Fragment(frag)
                frag = []
            
            frag.append(t)
            
        if frag:
            yield Fragment(frag)


class SentenceFragmenter(object):
    """Breaks the text up on sentence end punctuation characters (".", "!", or "?").
    This object works by looking in the original text for a sentence end as the next
    character after each token's 'endchar'.
    """
    
    def __init__(self, maxchars = 200, sentencechars = ".!?"):
        """
        :param maxchars: The maximum number of characters allowed in a fragment.
        """
        
        self.maxchars = maxchars
        self.sentencechars = frozenset(sentencechars)
    
    def __call__(self, text, tokens):
        maxchars = self.maxchars
        sentencechars = self.sentencechars
        textlen = len(text)
        first = None
        frag = []
        
        for t in tokens:
            if first is None:
                first = t.startchar
            endchar = t.endchar
            
            if endchar - first > maxchars:
                first = None
                if frag:
                    yield Fragment(frag)
                frag = []
            
            frag.append(t)
            if frag and endchar < textlen and text[endchar] in sentencechars:
                # Don't break for two periods in a row (e.g. ignore "...")
                if endchar+1 < textlen and text[endchar + 1] in sentencechars:
                    continue
                
                yield Fragment(frag, charsafter = 1)
                frag = []
                first = None
        
        if frag:
            yield Fragment(frag)


class ContextFragmenter(object):
    """Looks for matched terms and aggregates them with their
    surrounding context.
    
    This fragmenter only yields fragments that contain matched terms.    
    """
    
    def __init__(self, termset, maxchars = 200, charsbefore = 20, charsafter = 20):
        """
        :param termset: A collection (probably a set or frozenset) containing the
            terms you want to match to token.text attributes.
        :param maxchars: The maximum number of characters allowed in a fragment.
        :param charsbefore: The number of extra characters of context to add before
            the first matched term.
        :param charsafter: The number of extra characters of context to add after
            the last matched term.
        """
        
        self.maxchars = maxchars
        self.charsbefore = charsbefore
        self.charsafter = charsafter
        
    def __call__(self, text, tokens):
        maxchars = self.maxchars
        charsbefore = self.charsbefore
        charsafter = self.charsafter
        
        current = []
        currentlen = 0
        countdown = -1
        for t in tokens:
            if t.matched:
                countdown = charsafter
            
            current.append(t)
            
            length = t.endchar - t.startchar
            currentlen += length
            
            if countdown >= 0:
                countdown -= length
                
                if countdown < 0 or currentlen >= maxchars:
                    yield Fragment(current)
                    current = []
                    currentlen = 0
            
            else:
                while current and currentlen > charsbefore:
                    t = current.pop(0)
                    currentlen -= t.endchar - t.startchar

        if countdown >= 0:
            yield Fragment(current)


#class VectorFragmenter(object):
#    def __init__(self, termmap, maxchars = 200, charsbefore = 20, charsafter = 20):
#        """
#        :param termmap: A dictionary mapping the terms you're looking for to
#            lists of either (posn, startchar, endchar) or
#            (posn, startchar, endchar, boost) tuples.
#        :param maxchars: The maximum number of characters allowed in a fragment.
#        :param charsbefore: The number of extra characters of context to add before
#            the first matched term.
#        :param charsafter: The number of extra characters of context to add after
#            the last matched term.
#        """
#        
#        self.termmap = termmap
#        self.maxchars = maxchars
#        self.charsbefore = charsbefore
#        self.charsafter = charsafter
#    
#    def __call__(self, text, tokens):
#        maxchars = self.maxchars
#        charsbefore = self.charsbefore
#        charsafter = self.charsafter
#        textlen = len(text)
#        
#        vfrags = []
#        for term, data in self.termmap.iteritems():
#            if len(data) == 3:
#                t = Token(startchar = data[1], endchar = data[2])
#            elif len(data) == 4:
#                t = Token(startchar = data[1], endchar = data[2], boost = data[3])
#            else:
#                raise ValueError(repr(data))
#            
#            newfrag = VFragment([t], charsbefore, charsafter, textlen)
#            added = False
#            
#            for vf in vfrags:
#                if vf.overlaps(newfrag) and vf.overlapped_length(newfrag) < maxchars:
#                    vf.merge(newfrag)
#                    added = True
#                    break


# Fragment scorers

def BasicFragmentScorer(f):
    # Add up the boosts for the matched terms in this passage
    score = sum(t.boost for t in f.matches)
    
    # Favor diversity: multiply score by the number of separate
    # terms matched
    score *= len(f.matched_terms) * 100
    
    return score


# Fragment sorters

def SCORE(fragment):
    "Sorts higher scored passages first."
    return None
def FIRST(fragment):
    "Sorts passages from earlier in the document first."
    return fragment.startchar
def LONGER(fragment):
    "Sorts longer passages first."
    return 0 - len(fragment)
def SHORTER(fragment):
    "Sort shorter passages first."
    return len(fragment)


# Formatters

class UppercaseFormatter(object):
    def __init__(self, between = "..."):
        self.between = between
        
    def _format_fragment(self, text, fragment):
        output = []
        index = fragment.startchar
        
        for t in fragment.matches:
            if t.startchar > index:
                output.append(text[index:t.startchar])
            
            ttxt = text[t.startchar:t.endchar]
            if t.matched: ttxt = ttxt.upper()
            output.append(ttxt)
            index = t.endchar
        
        output.append(text[index:fragment.endchar])
        return "".join(output)

    def __call__(self, text, fragments):
        return self.between.join((self._format_fragment(text, fragment)
                                  for fragment in fragments))


class GenshiFormatter(object):
    def __init__(self, qname, between = "..."):
        self.qname = qname
        self.between = between
        
        from genshi.core import START, END, TEXT, Attrs, Stream #@UnresolvedImport
        self.START, self.END, self.TEXT, self.Attrs, self.Stream = (START, END, TEXT, Attrs, Stream)

    def _add_text(self, text, output):
        if output and output[-1][0] == self.TEXT:
            output[-1] = (self.TEXT, output[-1][1] + text, output[-1][2])
        else:
            output.append((self.TEXT, text, (None, -1, -1)))

    def _format_fragment(self, text, fragment):
        START, TEXT, END, Attrs = self.START, self.TEXT, self.END, self.Attrs
        qname = self.qname
        output = []
        
        index = fragment.startchar
        lastmatched = False
        for t in fragment.matches:
            if t.startchar > index:
                if lastmatched:
                    output.append((END, qname, (None, -1, -1)))
                    lastmatched = False
                self._add_text(text[index:t.startchar], output)
            
            ttxt = text[t.startchar:t.endchar]
            if not lastmatched:
                output.append((START, (qname, Attrs()), (None, -1, -1)))
                lastmatched = True
            output.append((TEXT, ttxt, (None, -1, -1)))
                                    
            index = t.endchar
        
        if lastmatched:
            output.append((END, qname, (None, -1, -1)))
        
        return output

    def __call__(self, text, fragments):
        output = []
        first = True
        for fragment in fragments:
            if not first:
                self._add_text(self.between, output)
            first = False
            output += self._format_fragment(text, fragment)
        
        return self.Stream(output)


# Highlighting

def top_fragments(text, terms, analyzer, fragmenter, top = 3,
                  scorer = BasicFragmentScorer, minscore = 1):
    termset = frozenset(terms)
    tokens = copyandmatchfilter(termset,
                                analyzer(text, chars = True, keeporiginal = True))
    
    scored_frags = nlargest(top, ((scorer(f), f) for f in fragmenter(text, tokens)))
    return [sf for score, sf in scored_frags if score > minscore]


def highlight(text, terms, analyzer, fragmenter, formatter, top=3,
              scorer = BasicFragmentScorer, minscore = 1,
              order = FIRST):
    
    fragments = top_fragments(text, terms, analyzer, fragmenter,
                              top = top, minscore = minscore)
    fragments.sort(key = order)
    return formatter(text, fragments)
    

if __name__ == '__main__':
    import re, time
    from whoosh import analysis
    #from genshi import QName
    
    sa = analysis.StemmingAnalyzer()
    txt = open("/Volumes/Storage/Development/help/documents/nodes/sop/copy.txt").read().decode("utf8")
    txt = re.sub("[\t\r\n ]+", " ", txt)
    t = time.time()
    fs = highlight(txt, ["templat", "geometri"], sa, SentenceFragmenter(), UppercaseFormatter())
    print time.time() - t
    print fs







import unittest
from os import mkdir
from os.path import exists
from shutil import rmtree

from whoosh import fields, index, qparser
from whoosh.filedb.filestore import FileStorage, RamStorage
from whoosh.filedb.filewriting import NO_MERGE


class TestIndexing(unittest.TestCase):
    def make_index(self, dirname, schema):
        if not exists(dirname):
            mkdir(dirname)
        st = FileStorage(dirname)
        ix = st.create_index(schema)
        return ix
    
    def destroy_index(self, dirname):
        if exists(dirname):
            rmtree(dirname)
    
    def test_creation(self):
        s = fields.Schema(content=fields.TEXT(phrase = True),
                          title=fields.TEXT(stored = True),
                          path=fields.ID(stored = True),
                          tags=fields.KEYWORD(stored = True),
                          quick=fields.NGRAM,
                          note=fields.STORED)
        st = RamStorage()
        
        ix = st.create_index(s)
        w = ix.writer()
        w.add_document(title=u"First", content=u"This is the first document", path=u"/a",
                       tags=u"first second third", quick=u"First document", note=u"This is the first document")
        w.add_document(content=u"Let's try this again", title=u"Second", path=u"/b",
                       tags=u"Uno Dos Tres", quick=u"Second document", note=u"This is the second document")
        w.commit()
        
    def test_integrity(self):
        s = fields.Schema(name = fields.TEXT, value = fields.TEXT)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(name = u"Yellow brown", value = u"Blue red green purple?")
        w.add_document(name = u"Alpha beta", value = u"Gamma delta epsilon omega.")
        w.commit()
        
        w = ix.writer()
        w.add_document(name = u"One two", value = u"Three four five.")
        w.commit()
        
        tr = ix.term_reader()
        self.assertEqual(ix.doc_count_all(), 3)
        self.assertEqual(list(tr.lexicon("name")), ["alpha", "beta", "brown", "one", "two", "yellow"])
    
    def test_lengths(self):
        s = fields.Schema(f1 = fields.KEYWORD(stored = True, scorable = True),
                          f2 = fields.KEYWORD(stored = True, scorable = True))
        ix = self.make_index("testindex", s)
        
        try:
            w = ix.writer()
            tokens = u"ABCDEFG"
            from itertools import cycle, islice
            lengths = [10, 20, 2, 102, 45, 3, 420, 2]
            for length in lengths:
                w.add_document(f2 = u" ".join(islice(cycle(tokens), length)))
            w.commit()
            dr = ix.doc_reader()
            ls1 = [dr.doc_field_length(i, "f1") for i in xrange(0, len(lengths))]
            ls2 = [dr.doc_field_length(i, "f2") for i in xrange(0, len(lengths))]
            self.assertEqual(ls1, [0]*len(lengths))
            self.assertEqual(ls2, lengths)
            dr.close()
            
            ix.close()
        finally:
            self.destroy_index("testindex")
    
    def test_lengths_ram(self):
        s = fields.Schema(f1 = fields.KEYWORD(stored = True, scorable = True),
                          f2 = fields.KEYWORD(stored = True, scorable = True))
        st = RamStorage()
        ix = st.create_index(s)
        w = ix.writer()
        w.add_document(f1 = u"A B C D E", f2 = u"X Y Z")
        w.add_document(f1 = u"B B B B C D D Q", f2 = u"Q R S T")
        w.add_document(f1 = u"D E F", f2 = u"U V A B C D E")
        w.commit()
        
        dr = ix.doc_reader()
        ls1 = [dr.doc_field_length(i, "f1") for i in xrange(0, 3)]
        ls2 = [dr.doc_field_length(i, "f2") for i in xrange(0, 3)]
        self.assertEqual(dr[0]["f1"], "A B C D E")
        self.assertEqual(dr.doc_field_length(0, "f1"), 5)
        self.assertEqual(dr.doc_field_length(1, "f1"), 8)
        self.assertEqual(dr.doc_field_length(2, "f1"), 3)
        self.assertEqual(dr.doc_field_length(0, "f2"), 3)
        self.assertEqual(dr.doc_field_length(1, "f2"), 4)
        self.assertEqual(dr.doc_field_length(2, "f2"), 7)
        
        self.assertEqual(ix.field_length("f1"), 16)
        self.assertEqual(ix.field_length("f2"), 14)
        
    def test_merged_lengths(self):
        s = fields.Schema(f1 = fields.KEYWORD(stored = True, scorable = True),
                          f2 = fields.KEYWORD(stored = True, scorable = True))
        st = RamStorage()
        ix = st.create_index(s)
        w = ix.writer()
        w.add_document(f1 = u"A B C", f2 = u"X")
        w.add_document(f1 = u"B C D E", f2 = u"Y Z")
        w.commit()
        
        w = ix.writer()
        w.add_document(f1 = u"A", f2 = u"B C D E X Y")
        w.add_document(f1 = u"B C", f2 = u"X")
        w.commit(NO_MERGE)
        
        w = ix.writer()
        w.add_document(f1 = u"A B X Y Z", f2 = u"B C")
        w.add_document(f1 = u"Y X", f2 = u"A B")
        w.commit(NO_MERGE)
        
        dr = ix.doc_reader()
        self.assertEqual(dr[0]["f1"], u"A B C")
        self.assertEqual(dr.doc_field_length(0, "f1"), 3)
        self.assertEqual(dr.doc_field_length(2, "f2"), 6)
        self.assertEqual(dr.doc_field_length(4, "f1"), 5)
        
    def test_frequency_keyword(self):
        s = fields.Schema(content = fields.KEYWORD)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(content = u"A B C D E")
        w.add_document(content = u"B B B B C D D")
        w.add_document(content = u"D E F")
        w.commit()
        
        tr = ix.term_reader()
        self.assertEqual(tr.doc_frequency("content", u"B"), 2)
        self.assertEqual(tr.frequency("content", u"B"), 5)
        self.assertEqual(tr.doc_frequency("content", u"E"), 2)
        self.assertEqual(tr.frequency("content", u"E"), 2)
        self.assertEqual(tr.doc_frequency("content", u"A"), 1)
        self.assertEqual(tr.frequency("content", u"A"), 1)
        self.assertEqual(tr.doc_frequency("content", u"D"), 3)
        self.assertEqual(tr.frequency("content", u"D"), 4)
        self.assertEqual(tr.doc_frequency("content", u"F"), 1)
        self.assertEqual(tr.frequency("content", u"F"), 1)
        self.assertEqual(tr.doc_frequency("content", u"Z"), 0)
        self.assertEqual(tr.frequency("content", u"Z"), 0)
        self.assertEqual(list(tr), [(0, u"A", 1, 1), (0, u"B", 2, 5),
                                    (0, u"C", 2, 2), (0, u"D", 3, 4),
                                    (0, u"E", 2, 2), (0, u"F", 1, 1)])
        
    def test_frequency_text(self):
        s = fields.Schema(content = fields.KEYWORD)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(content = u"alfa bravo charlie delta echo")
        w.add_document(content = u"bravo bravo bravo bravo charlie delta delta")
        w.add_document(content = u"delta echo foxtrot")
        w.commit()
        
        tr = ix.term_reader()
        self.assertEqual(tr.doc_frequency("content", u"bravo"), 2)
        self.assertEqual(tr.frequency("content", u"bravo"), 5)
        self.assertEqual(tr.doc_frequency("content", u"echo"), 2)
        self.assertEqual(tr.frequency("content", u"echo"), 2)
        self.assertEqual(tr.doc_frequency("content", u"alfa"), 1)
        self.assertEqual(tr.frequency("content", u"alfa"), 1)
        self.assertEqual(tr.doc_frequency("content", u"delta"), 3)
        self.assertEqual(tr.frequency("content", u"delta"), 4)
        self.assertEqual(tr.doc_frequency("content", u"foxtrot"), 1)
        self.assertEqual(tr.frequency("content", u"foxtrot"), 1)
        self.assertEqual(tr.doc_frequency("content", u"zulu"), 0)
        self.assertEqual(tr.frequency("content", u"zulu"), 0)
        self.assertEqual(list(tr), [(0, u"alfa", 1, 1), (0, u"bravo", 2, 5),
                                    (0, u"charlie", 2, 2), (0, u"delta", 3, 4),
                                    (0, u"echo", 2, 2), (0, u"foxtrot", 1, 1)])
    
    def test_deletion(self):
        s = fields.Schema(key = fields.ID, name = fields.TEXT, value = fields.TEXT)
        st = RamStorage()
        ix = st.create_index(s)
        
        w = ix.writer()
        w.add_document(key = u"A", name = u"Yellow brown", value = u"Blue red green purple?")
        w.add_document(key = u"B", name = u"Alpha beta", value = u"Gamma delta epsilon omega.")
        w.add_document(key = u"C", name = u"One two", value = u"Three four five.")
        w.commit()
        
        count = ix.delete_by_term("key", u"B")
        self.assertEqual(count, 1)
        ix.commit()
        
        self.assertEqual(ix.doc_count_all(), 3)
        self.assertEqual(ix.doc_count(), 2)
        
        ix.optimize()
        self.assertEqual(ix.doc_count(), 2)
        tr = ix.term_reader()
        self.assertEqual(list(tr.lexicon("name")), ["brown", "one", "two", "yellow"])

    def test_update(self):
        # Test update with multiple unique keys
        SAMPLE_DOCS = [{"id": u"test1", "path": u"/test/1", "text": u"Hello"},
                       {"id": u"test2", "path": u"/test/2", "text": u"There"},
                       {"id": u"test3", "path": u"/test/3", "text": u"Reader"},
                       ]
        
        schema = fields.Schema(id=fields.ID(unique=True, stored=True),
                               path=fields.ID(unique=True, stored=True),
                               text=fields.TEXT)
        ix = self.make_index("testindex", schema)
        try:
            writer = ix.writer()
            for doc in SAMPLE_DOCS:
                writer.add_document(**doc)
            writer.commit()
            
            writer = ix.writer()
            writer.update_document(**{"id": u"test2",
                                      "path": u"test/1",
                                      "text": u"Replacement"})
            writer.commit()
            ix.close()
        finally:
            self.destroy_index("testindex")

    def test_reindex(self):
        SAMPLE_DOCS = [
            {'id': u'test1', 'text': u'This is a document. Awesome, is it not?'},
            {'id': u'test2', 'text': u'Another document. Astounding!'},
            {'id': u'test3', 'text': u'A fascinating article on the behavior of domestic steak knives.'},
        ]

        schema = fields.Schema(text=fields.TEXT(stored=True),
                               id=fields.ID(unique=True, stored=True))
        ix = self.make_index("testindex", schema)
        try:
            def reindex():
                writer = ix.writer()
            
                for doc in SAMPLE_DOCS:
                    writer.update_document(**doc)
            
                writer.commit()

            reindex()
            self.assertEqual(ix.doc_count_all(), 3)
            reindex()
            self.assertEqual(ix.doc_count_all(), 3)
            
            ix.close()
            
        finally:
            self.destroy_index("testindex")



if __name__ == '__main__':
    unittest.main()



import unittest
from os import mkdir
from os.path import exists
from shutil import rmtree

from whoosh import analysis, fields, index, qparser
from whoosh.filedb.filestore import FileStorage
from whoosh.filedb import filetables


class TestTables(unittest.TestCase):
    def make_dir(self, dirname):
        if not exists(dirname):
            mkdir(dirname)
    
    def destroy_dir(self, dirname):
        if exists(dirname):
            rmtree(dirname)
    
    def test_hash(self):
        self.make_dir("testindex")
        st = FileStorage("testindex")
        hw = st.create_hash("test.hsh")
        hw.add("foo", "bar")
        hw.add("glonk", "baz")
        hw.close()
        
        hr = st.open_hash("test.hsh")
        self.assertEqual(hr.get("foo"), "bar")
        hr.close()
        
        self.destroy_dir("testindex")
    
    def test_posting_table(self):
        self.make_dir("testindex")
        st = FileStorage("testindex")
        ptw = st.create_posting_table("test.tiz", "test.pst",
                                     keycoder=filetables.encode_key,
                                     valuecoder=filetables.enpickle)
        f = fields.Frequency(None)
        writefn = f.write_postvalue
        readfn = f.read_postvalue
        
        for docnum, freq in [(1,2), (3,4), (5,6), (7,8)]:
            ptw.write_posting(docnum, freq, writefn)
        ptw.add((0, "foo"), "bar")
        
        for docnum, freq in [(10,20), (30,40), (50,60), (70,80)]:
            ptw.write_posting(docnum, freq, writefn)
        ptw.add((0, "glonk"), "baz")
        ptw.close()
        
        ptr = st.open_posting_table("test.tiz", "test.pst",
                                    keycoder=filetables.encode_key,
                                    keydecoder=filetables.decode_key,
                                    valuedecoder=filetables.depickle)
        self.assertEqual(ptr.get((0, "foo")), "bar")
        self.assertEqual(list(ptr.postings((0, "foo"), readfn)),
                         [(1,2), (3,4), (5,6), (7,8)])
        ptr.close()
        
        self.destroy_dir("testindex")


if __name__ == '__main__':
    unittest.main()



import unittest
from os import mkdir
from os.path import exists
from shutil import rmtree

from whoosh import analysis, fields, index, qparser
from whoosh.filedb.filestore import FileStorage


class TestVectors(unittest.TestCase):
    def make_index(self, dirname, schema):
        if not exists(dirname):
            mkdir(dirname)
        st = FileStorage(dirname)
        ix = st.create_index(schema)
        return ix
    
    def destroy_index(self, dirname):
        if exists(dirname):
            rmtree(dirname)
    
    def test_vector_merge(self):
        a = analysis.StandardAnalyzer()
        schema = fields.Schema(title = fields.TEXT,
                               content = fields.TEXT(vector=fields.Frequency(analyzer=a)))
        ix = self.make_index("testindex", schema)
        try:
            writer = ix.writer()
            writer.add_document(title=u"one",
                                content=u"This is the story of the black hole story")
            writer.commit()
            
            writer = ix.writer()
            writer.add_document(title=u"two",
                                content=u"You can read along in your book")
            writer.commit()
            
            searcher = ix.searcher()
            docnum = searcher.document_number(title=u"one")
            vec = list(searcher.vector(docnum, "content"))
            self.assertEqual(vec, [(u'black', 1), (u'hole', 1), (u'story', 2)])
            
            docnum = searcher.document_number(title=u"two")
            vec = list(searcher.vector(docnum, "content"))
            self.assertEqual(vec, [(u'along', 1), (u'book', 1), (u'read', 1), (u'your', 1)])
        finally:
            pass
            #self.destroy_index("testindex")


if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import mmap, os
from array import array
from cPickle import dump as dump_pickle
from cPickle import load as load_pickle
from marshal import dump as dump_marshal
from marshal import load as load_marshal
from struct import calcsize, pack, unpack
from sys import byteorder as _sys_byteorder


_SIZEMAP = dict((typecode, calcsize(typecode)) for typecode in "bBiIhHlLf")
_INT_SIZE = calcsize("i")
_USHORT_SIZE = calcsize("H")
_ULONG_SIZE = calcsize("L")
_FLOAT_SIZE = calcsize("f")
_ORDERMAP = {"little": "<", "big": ">"}


# Main function

def StructFile(fileobj, name=None, onclose=None, byteorder=None, mapped=True):
    """Returns a "structured file" object that wraps the given file object and provides
    numerous additional methods for writing structured data, such as "write_varint"
    and "write_ulong".
    
    The is a function which returns different classes based on (a) whether 'fileobj'
    is a real file or a "file-like object" (based on whether it has a fileno() method),
    and (b) whether 'byteorder' matches the native byte order.
    
    The following table lists the three different sub-classes of FileBase that may be
    returned:
    
    ------------ ------------------ ------------------
    Real file?   Native order       Non-native order
    ------------ ------------------ ------------------
    Yes          NativeRealFile     NonnativeFile
    No           NativeFilelike     NonnativeFile
    ------------ ------------------ ------------------
    
    :param fileobj: the file object to wrap.
    :param name: a name associated with the file object (for debugging purposes).
    :param onclose: a callback function to be called when close() is called on
        the StructFile object.
    :param byteorder: the byte order to read/write, either "big" or "little". If
        this parameter is not given, the native byte order is used.
    :param mapped: whether to use memory mapping. This value is only taken into
        account when (a) reading, (b) using the native byte-order, and (c) using
        a real file. Reading all other file types, always simulate memory mapping
        using seek and read.
    :rtype: :class:`FileBase`
    """
    
    if byteorder is None: byteorder = _sys_byteorder
    native = byteorder == _sys_byteorder
    
    if native:
        realfile = hasattr(fileobj, "fileno")
        if realfile:
            return NativeRealFile(fileobj, name=name, onclose=onclose, mapped=mapped)
        else:
            return NativeFilelike(fileobj, name=name, onclose=onclose)
    else:
        return NonnativeFile(fileobj, name=name, onclose=onclose,
                             orderchar=_ORDERMAP[byteorder])


# Utility functions

def float_to_byte(value, mantissabits = 5, zeroexp = 2):
    # Assume int size == float size
    
    fzero = (63 - zeroexp) << mantissabits
    bits = unpack("i", pack("f", value))[0]
    smallfloat = bits >> (24 - mantissabits)
    if smallfloat < fzero:
        # Map negative numbers and 0 to 0
        # Map underflow to next smallest non-zero number
        if bits <= 0:
            return 0
        else:
            return 1
    elif smallfloat >= fzero + 0x100:
        # Map overflow to largest number
        return 255
    else:
        return smallfloat - fzero
    
def byte_to_float(b, mantissabits = 5, zeroexp = 2):
    if b == 0:
        return 0.0
    
    bits = (b & 0xff) << (24 - mantissabits)
    bits += (63 - zeroexp) << 24
    return unpack("f", pack("i", bits))[0]
    

# Varint cache

# Build a cache of the varint byte sequences for the first
# N integers, so we don't have to constantly recalculate them
# on the fly. This makes a small but noticeable difference.

def encode_varint(i):
    s = ""
    while (i & ~0x7F) != 0:
        s += chr((i & 0x7F) | 0x80)
        i = i >> 7
    s += chr(i)
    return s

_varint_cache_size = 512
_varint_cache = []
for i in xrange(0, _varint_cache_size):
    _varint_cache.append(encode_varint(i))
_varint_cache = tuple(_varint_cache)


# Classes

class FileBase(object):
    def __init__(self, fileobj, name=None, onclose=None):
        self.file = fileobj
        self._name = name
        self.onclose = onclose
        self.is_closed = False
        
        for attr in ("read", "write", "tell", "seek"):
            if hasattr(fileobj, attr):
                setattr(self, attr, getattr(fileobj, attr))
                
        self._setup_fake_map()
        
    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self._name)
    
    def __del__(self):
        if not self.is_closed:
            self.close()

    def flush(self):
        """Flushes the buffer of the wrapped file. This is a no-op
        if the wrapped file does not have a flush method.
        """
        if hasattr(self.file, "flush"):
            self.file.flush()

    def close(self):
        """Closes the wrapped file. This is a no-op
        if the wrapped file does not have a close method.
        """
        
        del self.map
        if self.onclose:
            self.onclose(self)
        if hasattr(self.file, "close"):
            self.file.close()
        self.is_closed = True
    
    def _setup_fake_map(self): 
        _self = self
        class fakemap(object):
            def __getitem__(self, slice):
                _self.seek(slice.start)
                return _self.read(slice.stop - slice.start)
        self.map = fakemap()
    
    def write_string(self, s):
        """Writes a string to the wrapped file. This method writes the
        length of the string first, so you can read the string back
        without having to know how long it was.
        """
        self.write_varint(len(s))
        self.file.write(s)
        
    def read_string(self):
        """Reads a string from the wrapped file.
        """
        return self.file.read(self.read_varint())
    
    def skip_string(self):
        l = self.read_varint()
        self.seek(l, 1)
    
    def write_varint(self, i):
        """Writes a variable-length integer to the wrapped file.
        """
        assert i >= 0
        if i < len(_varint_cache):
            self.file.write(_varint_cache[i])
            return
        s = ""
        while (i & ~0x7F) != 0:
            s += chr((i & 0x7F) | 0x80)
            i = i >> 7
        s += chr(i)
        self.file.write(s)
        
    def read_varint(self):
        """Reads a variable-length encoded integer from the wrapped
        file.
        """
        read = self.read_byte
        b = read()
        i = b & 0x7F

        shift = 7
        while b & 0x80 != 0:
            b = read()
            i |= (b & 0x7F) << shift
            shift += 7
        return i
    
    def write_byte(self, n):
        """Writes a single byte to the wrapped file, shortcut for
        file.write(chr(n)).
        """
        self.file.write(chr(n))
        
    def read_byte(self):
        return ord(self.file.read(1))
    
    def get_byte(self, position):
        return ord(self.map[position:position+1])
    
    def write_8bitfloat(self, f, mantissabits = 5, zeroexp = 2):
        """Writes a byte-sized representation of floating point value
        f to the wrapped file.
        mantissabits is the number of bits to use for the mantissa
        (with the rest used for the exponent).
        zeroexp is the zero point for the exponent.
        """
        
        self.write_byte(float_to_byte(f, mantissabits, zeroexp))
    
    def read_8bitfloat(self, mantissabits = 5, zeroexp = 2):
        """Reads a byte-sized representation of a floating point value.
        mantissabits is the number of bits to use for the mantissa
        (with the rest used for the exponent).
        zeroexp is the zero point for the exponent.
        """
        return byte_to_float(self.read_byte(), mantissabits, zeroexp)
    
    def write_pickle(self, obj, protocol=-1):
        """Writes a pickled representation of obj to the wrapped file.
        """
        dump_pickle(obj, self.file, protocol)
    
    def write_marshal(self, obj):
        dump_marshal(obj, self.file)
    
    def read_pickle(self):
        """Reads a pickled object from the wrapped file.
        """
        return load_pickle(self.file)
    
    def read_marshal(self):
        return load_marshal(self.file)
    
    def write_sbyte(self, n):
        """Writes a signed byte value to the wrapped file.
        """
        raise NotImplementedError
    def write_int(self, n):
        """Writes a binary integer value to the wrapped file,.
        """
        raise NotImplementedError
    def write_ushort(self, n):
        """Writes an unsigned binary short integer value to the wrapped file.
        """
        raise NotImplementedError
    def write_ulong(self, n):
        """Writes an unsigned binary integer value to the wrapped file.
        """
        raise NotImplementedError
    def write_float(self, n):
        """Writes a binary float value to the wrapped file.
        """
        raise NotImplementedError
    def write_array(self, arry):
        """Writes an array to the wrapped file.
        """
        raise NotImplementedError
    def read_sbyte(self):
        """Reads a signed byte value from the wrapped file.
        """
        raise NotImplementedError
    def read_int(self):
        """Reads a binary integer value from the wrapped file.
        """
        raise NotImplementedError
    def read_ushort(self):
        """Reads an unsigned binary short integer value from the wrapped file.
        """
        raise NotImplementedError
    def read_ulong(self):
        """Reads an unsigned binary integer value from the wrapped file.
        """
        raise NotImplementedError
    def read_float(self):
        """Reads a binary floating point value from the wrapped file.
        """
        raise NotImplementedError
    def read_array(self, typecode, length):
        """Reads an array of 'length' items from the wrapped file.
        """
        raise NotImplementedError


class NativeBase(FileBase):
    def __init__(self, *args, **kwargs):
        super(NativeBase, self).__init__(*args, **kwargs)
        
        self.sbyte_array = array("b", [0])
        self.int_array = array("i", [0])
        self.ushort_array = array("H", [0])
        self.ulong_array = array("L", [0])
        self.float_array = array("f", [0.0])
    
    def get_sbyte(self, position):
        return unpack("=B", self.map[position:position+1])[0]
    def get_int(self, position):
        return unpack("=i", self.map[position:position+_INT_SIZE])[0]
    def get_ushort(self, position):
        return unpack("=H", self.map[position:position+_USHORT_SIZE])[0]
    def get_ulong(self, position):
        return unpack("=L", self.map[position:position+_ULONG_SIZE])[0]
    def get_float(self, position):
        return unpack("=f", self.map[position:position+_FLOAT_SIZE])[0]
    def get_array(self, position, typecode, length):
        return array(typecode,
                     unpack("=" + typecode * length,
                            self.map[position:position+_SIZEMAP[typecode] * length]))
        

class NativeRealFile(NativeBase):
    def __init__(self, fileobj, mapped=True, *args, **kwargs):
        super(NativeRealFile, self).__init__(fileobj, *args, **kwargs)
        
        # If mapped is True, set the 'map' attribute to a memory-mapped
        # representation of the file. Otherwise, the fake 'map' that
        # set up by the base class will be used.
        if mapped and "r" in fileobj.mode:
            fd = fileobj.fileno()
            self.size = os.fstat(fd).st_size
            self.map = mmap.mmap(fd, self.size, access=mmap.ACCESS_READ)
    
    def write_sbyte(self, n):
        self.sbyte_array[0] = n
        self.sbyte_array.tofile(self.file)
    def write_int(self, n):
        self.int_array[0] = n
        self.int_array.tofile(self.file)
    def write_ushort(self, n):
        self.ushort_array[0] = n
        self.ushort_array.tofile(self.file)
    def write_ulong(self, n):
        self.ulong_array[0] = n
        self.ulong_array.tofile(self.file)
    def write_float(self, n):
        self.float_array[0] = n
        self.float_array.tofile(self.file)
    def write_array(self, arry):
        arry.tofile(self.file)
        
    def read_sbyte(self):
        self.sbyte_array.fromfile(self.file, 1)
        return self.sbyte_array.pop()
    def read_int(self):
        self.int_array.fromfile(self.file, 1)
        return self.int_array.pop()
    def read_ushort(self):
        self.ushort_array.fromfile(self.file, 1)
        return self.ushort_array.pop()
    def read_ulong(self):
        self.ulong_array.fromfile(self.file, 1)
        return self.ulong_array.pop()
    def read_float(self):
        self.float_array.fromfile(self.file, 1)
        return self.float_array.pop()
    def read_array(self, typecode, length):
        arry = array(typecode)
        arry.fromfile(self.file, length)
        return arry


class NativeFilelike(NativeBase):
    def write_sbyte(self, n):
        self.sbyte_array[0] = n
        self.file.write(self.sbyte_array.tostring())
    def write_int(self, n):
        self.int_array[0] = n
        self.file.write(self.int_array.tostring())
    def write_ushort(self, n):
        self.ushort_array[0] = n
        self.file.write(self.ushort_array.tostring())
    def write_ulong(self, n):
        self.ulong_array[0] = n
        self.file.write(self.ulong_array.tostring())
    def write_float(self, n):
        self.float_array[0] = n
        self.file.write(self.float_array.tostring())
    def write_array(self, arry):
        self.file.write(arry.tostring())
    
    def read_sbyte(self):
        return unpack("=B", self.file.read(1))
    def read_int(self):
        return unpack("=i", self.file.read(_INT_SIZE))
    def read_ushort(self):
        return unpack("=H", self.file.read(_USHORT_SIZE))
    def read_ulong(self):
        return unpack("=L", self.file.read(_ULONG_SIZE))
    def read_float(self):
        return unpack("=f", self.file.read(_FLOAT_SIZE))
    def read_array(self, typecode, length):
        arry = array(typecode)
        arry.fromstring(self.file.read(arry.itemsize * length))
        return arry


class NonnativeFile(FileBase):
    def __init__(self,  orderchar="<", *args, **kwargs):
        super(NativeBase, self).__init__(*args, **kwargs)
        self.orderchar = orderchar
    
    def write_sbyte(self, n):
        self.file.write(pack(self.orderchar + "B", n))
    def write_int(self, n):
        self.file.write(pack(self.orderchar + "i", n))
    def write_ushort(self, n):
        self.file.write(pack(self.orderchar + "H", n))
    def write_ulong(self, n):
        self.file.write(pack(self.orderchar + "L", n))
    def write_float(self, n):
        self.file.write(pack(self.orderchar + "f", n))
    def write_array(self, arry):
        a = pack(self.orderchar + arry.typecode * len(arry), *arry)
        self.file.write(a)
        
    def read_sbyte(self):
        return unpack(self.orderchar + "B", self.file.read(1))
    def read_int(self):
        return unpack(self.orderchar + "i", self.file.read(_INT_SIZE))
    def read_ushort(self):
        return unpack(self.orderchar + "H", self.file.read(_USHORT_SIZE))
    def read_ulong(self):
        return unpack(self.orderchar + "L", self.file.read(_ULONG_SIZE))
    def read_float(self):
        return unpack(self.orderchar + "f", self.file.read(_FLOAT_SIZE))
    def read_array(self, typecode, length):
        return array(typecode,
                     unpack(self.orderchar + typecode * length,
                            self.file.read(_SIZEMAP[typecode] * length)))

    def get_sbyte(self, position):
        return unpack(self.orderchar + "B", self.map[position:position+1])[0]
    def get_int(self, position):
        return unpack(self.orderchar + "i", self.map[position:position+_INT_SIZE])[0]
    def get_ushort(self, position):
        return unpack(self.orderchar + "H", self.map[position:position+_USHORT_SIZE])[0]
    def get_ulong(self, position):
        return unpack(self.orderchar + "L", self.map[position:position+_ULONG_SIZE])[0]
    def get_float(self, position):
        return unpack(self.orderchar + "f", self.map[position:position+_FLOAT_SIZE])[0]
    def get_array(self, position, typecode, length):
        return array(typecode,
                     unpack(self.orderchar + typecode * length,
                            self.map[position:position+_SIZEMAP[typecode] * length]))










